from abc import abstractmethod
import re

from Corpus.Sentence import Sentence
from Dictionary.Word import Word
from Language.Language import Language


class SentenceSplitter:
    SEPARATORS = "\n()[]{}\"'\u05F4\uFF02\u055B’”‘“–­​	&  ﻿"
    SENTENCE_ENDERS = ".?!…"
    PUNCTUATION_CHARACTERS = ",:;‚"
    APOSTROPHES = "'’‘\u055B"

    @abstractmethod
    def upperCaseLetters(self) -> str:
        pass

    @abstractmethod
    def lowerCaseLetters(self) -> str:
        pass

    @abstractmethod
    def shortCuts(self) -> list:
        pass

    def __listContains(self, currentWord: str) -> bool:
        """
        The listContains method has a String array shortcuts which holds the possible abbreviations that might end with
        a '.' but not a sentence finisher word. It also takes a String as an input and loops through the shortcuts array
        and returns true if given String has any matching item in the shortcuts array.

        PARAMETERS
        ----------
        currentWord : str
            String input to check.

        RETURNS
        -------
        bool
            True if contains any abbreviations, False otherwise.
        """
        return currentWord in self.shortCuts()

    def __isNextCharUpperCaseOrDigit(self,
                                     line: str,
                                     i: int) -> bool:
        """
        The isNextCharUpperCaseOrDigit method takes a String line and an int i as inputs. First it compares each char in
        the input line with " " and SEPARATORS ({@literal ()[]{}"'״＂՛}) and increment i by one until a mismatch or end
        of line.

        When i equals to line length or contains one of the uppercase letters or digits it returns true, false
        otherwise.

        PARAMETERS
        ----------
        line : str
            String to check.
        i : int
            defining starting index.

        RETURNS
        -------
        bool
            True if next char is uppercase or digit, False otherwise.
        """
        while i < len(line) and (line[i] == ' ' or line[i] in SentenceSplitter.SEPARATORS):
            i = i + 1
        if i == len(line) or line[i] in self.upperCaseLetters() + Language.DIGITS + "-":
            return True
        else:
            return False

    def __isPreviousWordUpperCase(self,
                                  line: str,
                                  i: int) -> bool:
        """
        The isPreviousWordUpperCase method takes a String line and an int i as inputs. First it compares each char in
        the input line with " " and checks each char whether they are lowercase letters or one of the qxw. And decrement
        input i by one till this condition is false.

        When i equals to -1 or contains one of the uppercase letters or one of the QXW it returns true, false otherwise.

        PARAMETERS
        ----------
        line : str
            String to check.
        i : int
            defining ending index.

        RETURNS
        -------
        bool
            True if previous char is uppercase or one of the QXW, False otherwise.
        """
        while i >= 0 and (line[i] == ' ' or line[i] in self.lowerCaseLetters() + "qxw"):
            i = i - 1
        if i == -1 or line[i] in self.upperCaseLetters() + "QWX":
            return True
        else:
            return False

    def __isNextCharUpperCase(self,
                              line: str,
                              i: int) -> bool:
        """
        The isNextCharUpperCase method takes a String line and an int i as inputs. First it compares each char in
        the input line with " " and increment i by one until a mismatch or end of line.

        When i equals to line length or contains one of the uppercase letters it returns true, false otherwise.

        PARAMETERS
        ----------
        line : str
            String to check.
        i : int
            defining starting index.

        RETURNS
        -------
        bool
            True if next char is uppercase, False otherwise.
        """
        while i < len(line) and line[i] == ' ':
            i = i + 1
        if i == len(line) or line[i] in self.upperCaseLetters() + "\"\'":
            return True
        else:
            return False

    def __isNameShortcut(self, currentWord: str) -> bool:
        """
        The isNameShortcut method takes a String word as an input. First, if the word length is 1, and currentWord
        contains UPPERCASE_LETTERS letters than it returns true.

        Secondly, if the length of the word is 3 (i.e it is a shortcut) and it has a '.' at its 1st index and
        currentWord's 2nd  index is an uppercase letter it also returns true. (Ex : m.A)

        PARAMETERS
        ----------
        currentWord : str
            String input to check whether it is a shortcut.

        RETURNS
        -------
        bool
            True if given input is a shortcut, False otherwise.
        """
        if len(currentWord) == 1 and currentWord in self.upperCaseLetters():
            return True
        if len(currentWord) == 3 and currentWord[1] == '.' and currentWord[2] in self.upperCaseLetters():
            return True
        return False

    def __repeatControl(self,
                        word: str,
                        exceptionMode: bool) -> str:
        """
        The repeatControl method takes a String word as an input, and a boolean exceptionMode and compress the
        repetitive chars. With the presence of exceptionMode it directly returns the given word. Then it declares a
        counter i and loops till the end of the given word. It compares the char at index i with the char at index (i+2)
        if they are equal then it compares the char at index i with the char at index (i+1) and increments i by one and
        returns concatenated result String with char at index i.

        PARAMETERS
        ----------
        word : str
            String input.
        exceptionMode : bool
            boolean input for exceptional cases.

        RETURNS
        -------
        str
            String result.
        """
        if exceptionMode:
            return word
        i = 0
        result = ""
        while i < len(word):
            if i < len(word) - 2 and word[i] == word[i + 1] and word[i] == word[i + 2]:
                while i < len(word) - 1 and word[i] == word[i + 1]:
                    i = i + 1
            result = result + word[i]
            i = i + 1
        return result

    def __isApostrophe(self,
                       line: str,
                       i: int) -> bool:
        """
        The isApostrophe method takes a String line and an integer i as inputs. Initially declares a String
        apostropheLetters which consists of abcçdefgğhıijklmnoöprsştuüvyzABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZ,
        âàáäãèéêëíîòóôûúqwxÂÈÉÊËÌÒÛQWX and  0123456789. Then, it returns true if the result of contains method which
        checks the existence of previous char and next char at apostropheLetters returns true, returns false otherwise.

        PARAMETERS
        ----------
        line : str
            String input to check.
        i : int
            index.

        RETURNS
        -------
        bool
            True if apostropheLetters contains previous char and next char, False otherwise.
        """
        apostrophe_letters = self.upperCaseLetters() + self.lowerCaseLetters() + Language.EXTENDED_LANGUAGE_CHARACTERS \
                             + Language.DIGITS
        if i > 0 and i + 1 < len(line):
            previous_char = line[i - 1]
            next_char = line[i + 1]
            return previous_char in apostrophe_letters and next_char in apostrophe_letters
        else:
            return False

    def __numberExistsBeforeAndAfter(self,
                                     line: str,
                                     i: int) -> bool:
        """
        The numberExistsBeforeAndAfter method takes a String line and an integer i as inputs. Then, it returns true if
        the result of contains method, which compares the previous char and next char with 0123456789, returns true and
        false otherwise.

        PARAMETERS
        ----------
        line : str
            String input to check.
        i : int
            index.

        RETURNS
        -------
        bool
            True if previous char and next char is a digit, False otherwise.
        """
        if i > 0 and i + 1 < len(line):
            previous_char = line[i - 1]
            next_char = line[i + 1]
            return previous_char in Language.DIGITS and next_char in Language.DIGITS
        else:
            return False

    def __isTime(self,
                 line: str,
                 i: int) -> bool:
        """
        The isTime method takes a String line and an integer i as inputs. Then, it returns true if
        the result of the contains method, which compares the previous char, next char and two next chars with
        0123456789, returns true and false otherwise.

        PARAMETERS
        ----------
        line : str
            String input to check.
        i : int
            index.

        RETURNS
        -------
        bool
            True if previous char, next char and two next chars are digit, False otherwise.
        """
        if i > 0 and i + 2 < len(line):
            previous_char = line[i - 1]
            next_char = line[i + 1]
            two_next_char = line[i + 2]
            return previous_char in Language.DIGITS and next_char in Language.DIGITS and \
                   two_next_char in Language.DIGITS
        else:
            return False

    def split(self, line: str) -> list:
        """
        The split method takes a String line as an input. Firstly it creates a new sentence as currentSentence a new
        list as sentences. Then loops till the end of the line and checks some conditions;
        If the char at ith index is a separator;

        ' : assigns currentWord as currentWord'
        { : increment the curlyBracketCount
        } : decrement the curlyBracketCount
        " : increment the specialQuotaCount
        " : decrement the specialQuotaCount
        ( : increment roundParenthesisCount
        ) : decrement roundParenthesisCount
        [ : increment bracketCount
        ] : decrement bracketCount
        " : assign quotaCount as 1- quotaCount
        ' : assign apostropheCount as 1- apostropheCount

        If the currentWord is not empty, it adds the currentWord after repeatControl to currentSentence.

        If the char at index i is " and  bracketCount, specialQuotaCount, curlyBracketCount, roundParenthesisCount, and
        quotaCount equal to 0 and also the next char is uppercase or digit, it adds currentSentence to sentences.

        If the char at ith index is a sentence ender;

        . and currentWord is www : assigns webMode as true. Ex: www.google.com
        . and currentWord is a digit or in web or e-mail modes : assigns currentWord as currentWord+char(i) Ex: 1.
        . and currentWord is a shortcut or an abbreviation : assigns currentWord as currentWord+char(i) and adds
        currentWord to currentSentence. Ex : bkz.
        ' and next char is uppercase or digit: add word to currentSentence as ' and add currentSentence to sentences.

        If the char at index i is ' ', i.e space, add word to currentSentence and assign "" to currentSentence.
        If the char at index i is -,  add word to currentSentence and add sentences when the wordCount of
        currentSentence greater than 0.

        If the char at ith index is a punctuation;
        : and if currentWord is "https" : assign webMode as true.
        , and there exists a number before and after : assign currentWord as currentWord+char(i) Ex: 1,2
        : and if line is a time : assign currentWord as currentWord+char(i) Ex: 12:14:24
        - and there exists a number before and after : assign currentWord as currentWord+char(i) Ex: 12-1
        {@literal @} : assign emailMode as true.

        PARAMETERS
        ----------
        line : str
            String input to split.

        RETURNS
        -------
        list
            sentences list which holds split line.
        """
        email_mode = False
        web_mode = False
        i = 0
        special_quota_count = 0
        round_parenthesis_count = 0
        bracket_count = 0
        curly_bracket_count = 0
        quota_count = 0
        apostrophe_count = 0
        current_sentence = Sentence()
        current_word = ""
        sentences = []
        while i < len(line):
            if line[i] in SentenceSplitter.SEPARATORS:
                if line[i] in SentenceSplitter.APOSTROPHES and current_word != "" and self.__isApostrophe(line, i):
                    current_word = current_word + line[i]
                else:
                    if current_word != "":
                        current_sentence.addWord(Word(self.__repeatControl(current_word, web_mode or email_mode)))
                    if line[i] != '\n':
                        current_sentence.addWord(Word("" + line[i]))
                    current_word = ""
                    if line[i] == '{':
                        curly_bracket_count = curly_bracket_count + 1
                    elif line[i] == '}':
                        curly_bracket_count = curly_bracket_count - 1
                    elif line[i] == '\uFF02':
                        special_quota_count = special_quota_count + 1
                    elif line[i] == '\u05F4':
                        special_quota_count = special_quota_count - 1
                    elif line[i] == '“':
                        special_quota_count = special_quota_count + 1
                    elif line[i] == '”':
                        special_quota_count = special_quota_count - 1
                    elif line[i] == '‘':
                        special_quota_count = special_quota_count + 1
                    elif line[i] == '’':
                        special_quota_count = special_quota_count - 1
                    elif line[i] == '(':
                        round_parenthesis_count = round_parenthesis_count + 1
                    elif line[i] == ')':
                        round_parenthesis_count = round_parenthesis_count - 1
                    elif line[i] == '[':
                        bracket_count = bracket_count + 1
                    elif line[i] == ']':
                        bracket_count = bracket_count - 1
                    elif line[i] == '"':
                        quota_count = 1 - quota_count
                    elif line[i] == '\'':
                        apostrophe_count = 1 - apostrophe_count
                    if line[
                        i] == '"' and bracket_count == 0 and special_quota_count == 0 and curly_bracket_count == 0 and \
                            round_parenthesis_count == 0 and quota_count == 0 and self.__isNextCharUpperCaseOrDigit(
                        line,
                        i + 1):
                        sentences.append(current_sentence)
                        current_sentence = Sentence()
            else:
                if line[i] in SentenceSplitter.SENTENCE_ENDERS:
                    if line[i] == '.' and current_word.lower() == "www":
                        web_mode = True
                    if line[i] == '.' and current_word != "" and (
                            web_mode or email_mode or (
                            line[i - 1] in Language.DIGITS and not self.__isNextCharUpperCaseOrDigit(line, i + 1))):
                        current_word = current_word + line[i]
                        current_sentence.addWord(Word(current_word))
                        current_word = ""
                    else:
                        if line[i] == '.' and (
                                self.__listContains(current_word) or self.__isNameShortcut(current_word)):
                            current_word = current_word + line[i]
                            current_sentence.addWord(Word(current_word))
                            current_word = ""
                        else:
                            if line[i] == '.' and self.__numberExistsBeforeAndAfter(line, i):
                                current_word = current_word + line[i]
                            else:
                                if current_word != "":
                                    current_sentence.addWord(
                                        Word(self.__repeatControl(current_word, web_mode or email_mode)))
                                current_word = "" + line[i]
                                i = i + 1
                                while i < len(line) and line[i] in SentenceSplitter.SENTENCE_ENDERS:
                                    i = i + 1
                                i = i - 1
                                current_sentence.addWord(Word(current_word))
                                if round_parenthesis_count == 0 and bracket_count == 0 and curly_bracket_count == 0 and \
                                        quota_count == 0:
                                    if i + 1 < len(line) and line[i + 1] == '\'' and apostrophe_count == 1 and \
                                            self.__isNextCharUpperCaseOrDigit(line, i + 2):
                                        current_sentence.addWord(Word("'"))
                                        i = i + 1
                                        sentences.append(current_sentence)
                                        current_sentence = Sentence()
                                    else:
                                        if i + 2 < len(line) and line[i + 1] == ' ' and line[i + 2] == '\'' and \
                                                apostrophe_count == 1 and self.__isNextCharUpperCaseOrDigit(line,
                                                                                                            i + 3):
                                            current_sentence.addWord(Word("'"))
                                            i += 2
                                            sentences.append(current_sentence)
                                            current_sentence = Sentence()
                                        else:
                                            if self.__isNextCharUpperCaseOrDigit(line, i + 1):
                                                sentences.append(current_sentence)
                                                current_sentence = Sentence()
                                current_word = ""
                else:
                    if line[i] == ' ':
                        email_mode = False
                        web_mode = False
                        if current_word != "":
                            current_sentence.addWord(Word(self.__repeatControl(current_word, web_mode or email_mode)))
                            current_word = ""
                    else:
                        if line[i] == '-' and not web_mode and round_parenthesis_count == 0 and \
                                self.__isNextCharUpperCase(line, i + 1) and \
                                not self.__isPreviousWordUpperCase(line, i - 1):
                            if current_word != "" and current_word not in Language.DIGITS:
                                current_sentence.addWord(
                                    Word(self.__repeatControl(current_word, web_mode or email_mode)))
                            if current_sentence.wordCount() > 0:
                                sentences.append(current_sentence)
                            current_sentence = Sentence()
                            round_parenthesis_count = 0
                            bracket_count = 0
                            curly_bracket_count = 0
                            quota_count = 0
                            special_quota_count = 0
                            if current_word != "" and re.match("\\d+", current_word):
                                current_sentence.addWord(Word(current_word + " -"))
                            else:
                                current_sentence.addWord(Word("-"))
                            current_word = ""
                        else:
                            if line[i] in SentenceSplitter.PUNCTUATION_CHARACTERS or \
                                    line[i] in Language.ARITHMETIC_CHARACTERS:
                                if line[i] == ':' and (current_word == "http" or current_word == "https"):
                                    web_mode = True
                                if web_mode:
                                    current_word = current_word + line[i]
                                else:
                                    if line[i] == ',' and self.__numberExistsBeforeAndAfter(line, i):
                                        current_word = current_word + line[i]
                                    else:
                                        if line[i] == ':' and self.__isTime(line, i):
                                            current_word = current_word + line[i]
                                        else:
                                            if line[i] == '-' and self.__numberExistsBeforeAndAfter(line, i):
                                                current_word = current_word + line[i]
                                            else:
                                                if current_word != "":
                                                    current_sentence.addWord(
                                                        Word(
                                                            self.__repeatControl(current_word, web_mode or email_mode)))
                                                current_sentence.addWord(Word("" + line[i]))
                                                current_word = ""
                            else:
                                if line[i] == '@':
                                    current_word = current_word + line[i]
                                    email_mode = True
                                else:
                                    current_word = current_word + line[i]
            i = i + 1
        if current_word != "":
            current_sentence.addWord(Word(self.__repeatControl(current_word, web_mode or email_mode)))
        if current_sentence.wordCount() > 0:
            sentences.append(current_sentence)
        return sentences



from Corpus.SentenceSplitter import SentenceSplitter
from Language.TurkishLanguage import TurkishLanguage


class TurkishSplitter(SentenceSplitter):

    def upperCaseLetters(self) -> str:
        return TurkishLanguage.UPPERCASE_LETTERS

    def lowerCaseLetters(self) -> str:
        return TurkishLanguage.LOWERCASE_LETTERS

    def shortCuts(self) -> list:
        return ["alb", "bnb", "bkz", "bşk", "co", "dr", "dç", "der", "em", "gn",
                "hz", "kd", "kur", "kuv", "ltd", "md", "mr", "mö", "muh", "müh",
                "no", "öğr", "op", "opr", "org", "sf", "tuğ", "uzm", "vb", "vd",
                "yön", "yrb", "yrd", "üniv", "fak", "prof", "dz", "yd", "krm", "gen",
                "pte", "p", "av", "II", "III", "IV", "VI", "VII", "VIII", "IX",
                "X", "XI", "XII", "XIII", "XIV", "XV", "XVI", "XVII", "XVIII", "XIX",
                "XX", "tuğa", "plt", "tğm", "tic", "srv", "bl", "dipl", "not", "min",
                "cul", "san", "rzv", "or", "kor", "tüm", "st", "sn", "fr", "pl",
                "ka", "tk", "ko", "vs", "yard", "bknz", "doç", "gör", "müz", "oyn",
                "m", "s", "kr", "ms", "hv", "uz", "re", "ph", "mc", "ed",
                "km", "yb", "bk", "jr", "bn", "os", "mrs", "bld", "sen", "alm",
                "sir", "ord", "dir", "yay", "man", "brm", "edt", "dec", "mah", "cad",
                "vol", "kom", "sok", "apt", "elk", "mad", "ort", "cap", "ste", "exc",
                "ef"]



from __future__ import annotations
from DataStructure.CounterHashMap import CounterHashMap
from Dictionary.Word import Word

from Corpus.AbstractCorpus import AbstractCorpus
from Corpus.LanguageChecker import LanguageChecker
from Corpus.Paragraph import Paragraph
from Corpus.Sentence import Sentence
from Corpus.SentenceSplitter import SentenceSplitter

import random


class Corpus(AbstractCorpus):

    paragraphs: list
    sentences: list
    word_list: CounterHashMap
    sentence_index: int

    def __init__(self,
                 fileName=None,
                 splitterOrChecker=None):
        """
        Constructor of Corpus class which takes a file name as an input. Then reads the input file line by line
        and calls addSentence method with each read line.

        PARAMETERS
        ----------
        fileName : str
            String file name input that will be read.
        """
        self.sentences = []
        self.paragraphs = []
        self.word_list = CounterHashMap()
        self.sentence_index = 0
        if fileName is not None:
            self.file_name = fileName
            file = open(self.file_name, "r", encoding='utf8')
            lines = file.readlines()
            file.close()
            if splitterOrChecker is not None:
                if isinstance(splitterOrChecker, SentenceSplitter):
                    for line in lines:
                        sentences = splitterOrChecker.split(line.strip())
                        paragraph = Paragraph()
                        for sentence in sentences:
                            paragraph.addSentence(sentence)
                        self.addParagraph(paragraph)
                elif isinstance(splitterOrChecker, LanguageChecker):
                    for line in lines:
                        sentence = Sentence(line.strip(), splitterOrChecker)
                        self.addSentence(sentence)
            else:
                for line in lines:
                    self.addSentence(Sentence(line.strip()))

    def combine(self, corpus: Corpus):
        """
        The combine method takes a Corpus as an input and adds each sentence of sentences list.

        PARAMETERS
        ----------
        corpus : Corpus
            Corpus type input.
        """
        for sentence in corpus.sentences:
            self.addSentence(sentence)

    def addSentence(self, s: Sentence):
        """
        The addSentence method takes a Sentence as an input. It adds given input to sentences list and loops
        through the each word in sentence and puts these words into wordList CounterHashMap.

        PARAMETERS
        ----------
        s : Sentence
            Sentence type input that will be added to sentences list and its words will be added to wordList
            CounterHashMap.
        """
        self.sentences.append(s)
        for i in range(s.wordCount()):
            w = s.getWord(i)
            self.word_list.put(w)

    def numberOfWords(self) -> int:
        """
        The numberOfWords method loops through the sentences list and accumulates the number of words in sentence.

        RETURNS
        -------
        int
            size which holds the total number of words.
        """
        size = 0
        for s in self.sentences:
            size += s.wordCount()
        return size

    def contains(self, word: str) -> bool:
        """
        The contains method takes a String word as an input and checks whether wordList CounterHashMap has the
        given word and returns true if so, otherwise returns false.

        PARAMETERS
        ----------
        word : str
            String input to check.

        RETURNS
        -------
        bool
            True if wordList has the given word, False otherwise.
        """
        return Word(word) in self.word_list

    def addParagraph(self, p: Paragraph):
        """
        The addParagraph method takes a Paragraph type input. It gets the sentences in the given paragraph and
        add these to the sentences list and the words in the sentences to the wordList CounterHashMap.

        PARAMETERS
        ----------
        p : Paragraph
            Paragraph type input to add sentences and wordList.
        """
        self.paragraphs.append(p)
        for i in range(p.sentenceCount()):
            self.addSentence(p.getSentence(i))

    def getFileName(self) -> str:
        """
        Getter for the file name.

        RETURNS
        -------
        str
            file name.
        """
        return self.file_name

    def getWordList(self) -> set:
        """
        Getter for the wordList.

        RETURNS
        -------
        set
            The keySet of wordList.
        """
        return set(self.word_list.keys())

    def wordCount(self) -> int:
        """
        The wordCount method returns the size of the wordList CounterHashMap.

        RETURNS
        -------
        int
            The size of the wordList CounterHashMap.
        """
        return len(self.word_list)

    def getCount(self, word: Word) -> int:
        """
        The getCount method returns the count value of given word.

        PARAMETERS
        ----------
        word : Word
            Word type input to check.

        RETURNS
        -------
        int
            The count value of given word.
        """
        return self.word_list[word]

    def sentenceCount(self) -> int:
        """
        The sentenceCount method returns the size of the sentences list.

        RETURNS
        -------
        int
            The size of the sentences list.
        """
        return len(self.sentences)

    def getSentence(self, index: int) -> Sentence:
        """
        Getter for getting a sentence at given index.

        PARAMETERS
        ----------
        index : int
            index to get sentence from.

        RETURNS
        -------
        Sentence
            The sentence at given index.
        """
        return self.sentences[index]

    def paragraphCount(self) -> int:
        """
        The paragraphCount method returns the size of the paragraphs list.

        RETURNS
        -------
        int
            The size of the paragraphs list.
        """
        return len(self.paragraphs)

    def getParagraph(self, index: int) -> Paragraph:
        """
        Getter for getting a paragraph at given index.

        PARAMETERS
        ----------
        index : int
            index to get paragraph from.

        RETURNS
        -------
        Paragraph
            The paragraph at given index.
        """
        return self.paragraphs[index]

    def maxSentenceLength(self) -> int:
        """
        The maxSentenceLength method finds the sentence with the maximum number of words and returns this number.

        RETURNS
        -------
        int
            maximum length.
        """
        max_length = 0
        for s in self.sentences:
            if s.wordCount() > max_length:
                max_length = s.wordCount()
        return max_length

    def getAllWordsAsList(self) -> list:
        """
        The getAllWordsAsList method creates new list of lists and adds each word in each sentence of sentences
        list into new list.

        RETURNS
        -------
        list
            Newly created and populated list.
        """
        all_words = []
        for i in range(self.sentenceCount()):
            all_words.append(self.getSentence(i).getWords())
        return all_words

    def shuffleSentences(self, seed: int):
        """
        The shuffleSentences method randomly shuffles sentences list with given seed value.

        PARAMETERS
        ----------
        seed : int
            value to randomize shuffling.
        """
        random.seed(seed)
        random.shuffle(self.sentences)

    def getTrainCorpus(self,
                       foldNo: int,
                       foldCount: int) -> Corpus:
        """
        The getTrainCorpus method takes two integer inputs foldNo and foldCount for determining train data size and
        count of fold respectively. Initially creates a new empty Corpus, then finds the sentenceCount as N. Then,
        starting from the index 0 it loops through the index (foldNo * N) / foldCount and add each sentence of sentences
        list to new Corpus. Later on, starting from the index ((foldNo + 1) * N) / foldCount, it loops through the index
        N and add each sentence of sentences list to new Corpus.

        PARAMETERS
        ----------
        foldNo : int
            Integer input for train set size.
        foldCount : int
            Integer input for counting fold.

        RETURNS
        -------
        Corpus
            The newly created and populated Corpus.
        """
        train_corpus = Corpus()
        N = self.sentenceCount()
        for i in range((foldNo * N) // foldCount):
            train_corpus.addSentence(self.sentences[i])
        for i in range(((foldNo + 1) * N) // foldCount, N):
            train_corpus.addSentence(self.sentences[i])
        return train_corpus

    def getTestCorpus(self,
                      foldNo: int,
                      foldCount: int) -> Corpus:
        """
        The getTestCorpus method takes two integer inputs foldNo and foldCount for determining test data size and count
        of fold respectively. Initially creates a new empty Corpus, then finds the sentenceCount as N. Then, starting
        from the index (foldNo * N) / foldCount it loops through the index ((foldNo + 1) * N) / foldCount and add each
        sentence of sentences list to new Corpus.

        PARAMETERS
        ----------
        foldNo : int
            Integer input for test size.
        foldCount : int
            Integer input counting fold.

        RETURNS
        -------
        Corpus
            The newly created and populated Corpus.
        """
        test_corpus = Corpus()
        N = self.sentenceCount()
        for i in range((foldNo * N) // foldCount, ((foldNo + 1) * N) // foldCount):
            test_corpus.addSentence(self.sentences[i])
        return test_corpus

    def __repr__(self):
        return f"{self.sentences}"

    def open(self):
        self.sentence_index = 0

    def close(self):
        self.sentence_index = 0

    def getNextSentence(self) -> Sentence:
        index = self.sentence_index
        if self.sentence_index < len(self.sentences):
            self.sentence_index = self.sentence_index + 1
            return self.sentences[index]
        else:
            return None



from typing import IO

from Corpus.AbstractCorpus import AbstractCorpus
from Corpus.Sentence import Sentence


class CorpusStream(AbstractCorpus):

    file: IO

    def __init__(self, fileName=None):
        self.file_name = fileName

    def open(self):
        self.file = open(self.file_name, "r", encoding='utf8')

    def close(self):
        self.file.close()

    def getNextSentence(self) -> Sentence:
        line = self.file.readline()
        if line:
            return Sentence(line)
        else:
            return None

    def getSentenceBatch(self, lineCount: int) -> list:
        sentences = []
        for i in range(lineCount):
            line = self.file.readline()
            if line:
                sentences.append(Sentence(line))
            else:
                break
        return sentences



from os.path import isfile


class FileDescription:

    __path: str
    __extension: str
    __index: int

    def __init__(self,
                 path: str,
                 extensionOrFileName: str,
                 index: int = None):
        self.__path = path
        if index is None:
            self.__extension = extensionOrFileName[extensionOrFileName.rindex('.') + 1:]
            self.__index = int(extensionOrFileName[0 : extensionOrFileName.rindex('.')])
        else:
            self.__extension = extensionOrFileName
            self.__index = index

    def getPath(self) -> str:
        return self.__path

    def getIndex(self) -> int:
        return self.__index

    def getExtension(self) -> str:
        return self.__extension

    def getFileName(self,
                    thisPath=None,
                    extension=None) -> str:
        if thisPath is None:
            thisPath = self.__path
        return self.getFileNameWithIndex(thisPath,
                                         self.__index,
                                         extension)

    def getFileNameWithExtension(self, extension: str) -> str:
        return self.getFileName(self.__path, extension)

    def getFileNameWithIndex(self,
                             thisPath: str,
                             index: int,
                             extension=None) -> str:
        if extension is None:
            extension = self.__extension
        return "%s/%04d.%s" % (thisPath, index, extension)

    def getRawFileName(self) -> str:
        return "%04d.%s" % (self.__index, self.__extension)

    def addToIndex(self, count: int):
        self.__index += count

    def nextFileExists(self,
                       count: int,
                       thisPath=None):
        if thisPath is None:
            thisPath = self.__path
        return isfile(self.getFileNameWithIndex(thisPath, self.__index + count))

    def previousFileExists(self,
                           count: int,
                           thisPath=None):
        if thisPath is None:
            thisPath = self.__path
        return isfile(self.getFileNameWithIndex(thisPath, self.__index - count))

    def __repr__(self):
        return f"{self.__path} {self.__index}.{self.__extension}"



from Corpus.SentenceSplitter import SentenceSplitter
from Language.EnglishLanguage import EnglishLanguage


class EnglishSplitter(SentenceSplitter):

    def upperCaseLetters(self) -> str:
        return EnglishLanguage.UPPERCASE_LETTERS

    def lowerCaseLetters(self) -> str:
        return EnglishLanguage.LOWERCASE_LETTERS

    def shortCuts(self) -> list:
        return ["dr", "prof", "org", "II", "III", "IV", "VI", "VII", "VIII", "IX",
                "X", "XI", "XII", "XIII", "XIV", "XV", "XVI", "XVII", "XVIII", "XIX",
                "XX", "min", "km", "jr", "mrs", "sir"]



from __future__ import annotations

import io

from Dictionary.Word import Word
from Corpus.LanguageChecker import LanguageChecker


class Sentence:

    words: list

    def __init__(self,
                 fileOrStr=None,
                 languageChecker: LanguageChecker = None):
        """
        Another constructor of Sentence class which takes a fileName as an input. It reads each word in the file
        and adds to words list.

        PARAMETERS
        ----------
        fileOrStr: str
            input file to read words from.
        """
        self.words = []
        if isinstance(fileOrStr, io.StringIO):
            lines = fileOrStr.readlines()
            for line in lines:
                word_list = line.split(" ")
                for word in word_list:
                    self.words.append(Word(word))
            fileOrStr.close()
        elif isinstance(fileOrStr, str):
            word_array = fileOrStr.split(" ")
            for word in word_array:
                if len(word) > 0:
                    if languageChecker is None or languageChecker.isValidWord(word):
                        self.words.append(Word(word))

    def __eq__(self, s: Sentence) -> bool:
        """
        The equals method takes a Sentence as an input. First compares the sizes of both words lists and words
        of the Sentence input. If they are not equal then it returns false. Than it compares each word in the list.
        If they are equal, it returns true.

        PARAMETERS
        ----------
        s : Sentence
            Sentence to compare.

        RETURNS
        -------
        bool
            True if words of two sentences are equal.
        """
        if len(self.words) != len(s.words):
            return False
        for i in range(len(self.words)):
            if self.words[i].getName() != s.words[i].getName():
                return False
        return True

    def getWord(self, index: int) -> Word:
        """
        The getWord method takes an index input and gets the word at that index.

        PARAMETERS
        ----------
        index : int
            is used to get the word.

        RETURNS
        -------
        Word
            the word in given index.
        """
        return self.words[index]

    def getWords(self) -> list:
        """
        The getWords method returns the words list.

        RETURNS
        -------
        list
            Words ArrayList.
        """
        return self.words

    def getStrings(self) -> list:
        """
        The getStrings method loops through the words list and adds each words' names to the newly created result list.

        RETURNS
        -------
        list
            Result list which holds names of the words.
        """
        result = []
        for word in self.words:
            result.append(word.getName())
        return result

    def getIndex(self, word: Word) -> int:
        """
        The getIndex method takes a word as an input and finds the index of that word in the words list if it exists.

        PARAMETERS
        ----------
        word : Word
            Word type input to search for.

        RETURNS
        -------
        int
            Index of the found input, -1 if not found.
        """
        return self.words.index(word)

    def wordCount(self) -> int:
        """
        The wordCount method finds the size of the words list.

        RETURNS
        -------
        int
            The size of the words list.
        """
        return len(self.words)

    def addWord(self, word: Word):
        """
        The addWord method takes a word as an input and adds this word to the words list.

        PARAMETERS
        ----------
        word : Word
            Word to add words list.
        """
        self.words.append(word)

    def charCount(self) -> int:
        """
        The charCount method finds the total number of chars in each word of words list.

        RETURNS
        -------
        int
            number of the chars in the whole sentence.
        """
        total = 0
        for word in self.words:
            total += word.charCount()
        return total

    def insertWord(self,
                   i: int,
                   newWord: Word):
        """
        The insertWord method takes an index and a word as inputs. It inserts the word at given index to words
        list.

        PARAMETERS
        ----------
        i : int
            index.
        newWord : Word
            to add the words list.
        """
        self.words.insert(i, newWord)

    def replaceWord(self,
                    i: int,
                    newWord: Word):
        """
        The replaceWord method takes an index and a word as inputs. It removes the word at given index from words
        list and then adds the given word to given index of words.

        PARAMETERS
        ----------
        i : int
            index.
        newWord : Word
            to add the words list.
        """
        self.words.pop(i)
        self.words.insert(i, newWord)

    def safeIndex(self, index: int) -> bool:
        """
        The safeIndex method takes an index as an input and checks whether this index is between 0 and the size of the
        words.

        PARAMETERS
        ----------
        index : int
            is used to check the safety.

        RETURNS
        -------
        bool
            true if an index is safe, false otherwise.
        """
        return 0 <= index < len(self.words)

    def __str__(self) -> str:
        """
        The overridden toString method returns an accumulated string of each word in words list.

        RETURNS
        -------
        str
            String result which has all the word in words list.
        """
        if len(self.words) > 0:
            result = self.words[0].__str__()
            for i in range(1, len(self.words)):
                result = result + " " + self.words[i].__str__()
            return result
        else:
            return ""

    def toString(self) -> str:
        """
        The toWords method returns an accumulated string of each word's names in words list.

        RETURNS
        -------
        str
            String result which has all the names of each item in words list.
        """
        if len(self.words) > 0:
            result = self.words[0].getName()
            for i in range(1, len(self.words)):
                result = result + " " + self.words[i].getName()
            return result
        else:
            return ""

    def writeToFile(self, fileName: str):
        """
        The writeToFile method writes the given file by using toString method.

        PARAMETERS
        ----------
        fileName : str
            file to write in.
        """
        out_file = open(fileName, "w", encoding="utf8")
        out_file.write(self.__str__() + "\n")
        out_file.close()

    def __repr__(self):
        return f"{self.words}"



from abc import abstractmethod
from Corpus.Sentence import Sentence


class AbstractCorpus:

    file_name: str

    @abstractmethod
    def open(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def getNextSentence(self) -> Sentence:
        pass



from Corpus.Sentence import Sentence


class Paragraph:

    __sentences: list

    def __init__(self):
        """
        A constructor of Paragraph class which creates a list sentences.
        """
        self.__sentences = []

    def addSentence(self, s: Sentence):
        """
        The addSentence method adds given sentence to sentences list.

        PARAMETERS
        ----------
        s : Sentence
            Sentence type input to add sentences.
        """
        self.__sentences.append(s)

    def sentenceCount(self) -> int:
        """
        The sentenceCount method finds the size of the list sentences.

        RETURNS
        -------
        int
            The size of the list sentences.
        """
        return len(self.__sentences)

    def getSentence(self, index: int) -> Sentence:
        """
        The getSentence method finds the sentence from sentences list at given index.

        PARAMETERS
        ----------
        index : int
            used to get a sentence.

        RETURNS
        -------
        Sentence
            sentence at given index.
        """
        return self.__sentences[index]



from Language.TurkishLanguage import TurkishLanguage

from Corpus.LanguageChecker import LanguageChecker
from Corpus.SentenceSplitter import SentenceSplitter


class TurkishChecker(LanguageChecker):

    def isValidWord(self, word: str) -> bool:
        """
        The isValidWord method takes an input String as a word than define all valid characters as a validCharacters
        string which has letters (abcçdefgğhıijklmnoöprsştuüvyzABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZ),
        extended language characters (âàáäãèéêëíîòóôûúqwxÂÈÉÊËÌÒÛQWX),
        digits (0123456789),
        separators ({@literal ()[]{}"'״＂՛}),
        sentence enders (.?!…),
        arithmetic chars (+-=*\\),
        punctuation chars (,:;),
        special-meaning chars

        Then, loops through input word's each char and if a char in word does not in the validCharacters string it
        returns false, true otherwise.

        PARAMETERS
        ----------
        word : str
            String to check validity.

        RETURNS
        -------
        bool
            True if each char in word is valid, False otherwise.
        """
        special_meaning_characters = "$\\_|@%#£§&><"
        valid_characters = TurkishLanguage.LETTERS + TurkishLanguage.EXTENDED_LANGUAGE_CHARACTERS + \
                          TurkishLanguage.DIGITS + SentenceSplitter.SEPARATORS + SentenceSplitter.SENTENCE_ENDERS + \
                          TurkishLanguage.ARITHMETIC_CHARACTERS + SentenceSplitter.PUNCTUATION_CHARACTERS + \
                          special_meaning_characters
        for i in range(len(word)):
            if word[i] not in valid_characters:
                return False
        return True
